@@ -1,7 +1,7 @@
 (*---------------------------------------------------------------------------
-   Copyright 2012 Daniel C. Bünzli. All rights reserved.
-   Distributed under the BSD3 license, see license at the end of the file.
-   uutf release 0.9.4
+   Copyright (c) 2012 Daniel C. Bünzli. All rights reserved.
+   Distributed under the ISC license, see terms at the end of the file.
+   uutf v1.0.2
   ---------------------------------------------------------------------------*)
 
 let io_buffer_size = 65536                           (* IO_BUFFER_SIZE 4.0.0 *)
@@ -16,26 +16,15 @@ let invalid_bounds j l =
    the module. He won't be upset. *)
 
 let unsafe_chr = Char.unsafe_chr
-let unsafe_blit = String.unsafe_blit
+let unsafe_blit = Bytes.unsafe_blit
 let unsafe_array_get = Array.unsafe_get
-let unsafe_byte s j = Char.code (String.unsafe_get s j)
-let unsafe_set_byte s j byte = String.unsafe_set s j (Char.unsafe_chr byte)
+let unsafe_byte s j = Char.code (Bytes.unsafe_get s j)
+let unsafe_set_byte s j byte = Bytes.unsafe_set s j (Char.unsafe_chr byte)
 
 (* Unicode characters *)
 
-type uchar = int
-let u_bom = 0xFEFF                                                   (* BOM. *)
-let u_rep = 0xFFFD                                 (* replacement character. *)
-let is_uchar cp =
-  (0x0000 <= cp && cp <= 0xD7FF) || (0xE000 <= cp && cp <= 0x10FFFF)
-
-let pp_cp ppf cp =
-  if cp < 0 || cp > 0x10FFFF then pp ppf "U+Invalid(%X)" cp else
-  if cp <= 0xFFFF then pp ppf "U+%04X" cp else
-  pp ppf "U+%X" cp
-
-let cp_to_string cp =                                    (* NOT thread safe. *)
-  pp Format.str_formatter "%a" pp_cp cp; Format.flush_str_formatter ()
+let u_bom = Uchar.unsafe_of_int 0xFEFF                               (* BOM. *)
+let u_rep = Uchar.unsafe_of_int 0xFFFD             (* replacement character. *)
 
 (* Unicode encoding schemes *)
 
@@ -62,23 +51,23 @@ let encoding_to_string = function
 
 (* Base character decoders. They assume enough data. *)
 
-let malformed s j l = `Malformed (String.sub s j l)
+let malformed s j l = `Malformed (Bytes.sub_string s j l)
 let malformed_pair be hi s j l =    (* missing or half low surrogate at eoi. *)
-  let bs1 = String.sub s j l in
-  let bs0 = String.create 2 in
+  let bs1 = Bytes.(sub s j l) in
+  let bs0 = Bytes.create 2 in
   let j0, j1 = if be then (0, 1) else (1, 0) in
   unsafe_set_byte bs0 j0 (hi lsr 8);
   unsafe_set_byte bs0 j1 (hi land 0xFF);
-  `Malformed (bs0 ^ bs1)
+  `Malformed Bytes.(unsafe_to_string (cat bs0 bs1))
 
 let r_us_ascii s j =
   (* assert (0 <= j && j < String.length s); *)
   let b0 = unsafe_byte s j in
-  if b0 <= 127 then `Uchar b0 else malformed s j 1
+  if b0 <= 127 then `Uchar (Uchar.unsafe_of_int b0) else malformed s j 1
 
 let r_iso_8859_1 s j =
   (* assert (0 <= j && j < String.length s); *)
-  `Uchar (unsafe_byte s j)
+  `Uchar (Uchar.unsafe_of_int @@ unsafe_byte s j)
 
 let utf_8_len = [| (* uchar byte length according to first UTF-8 byte. *)
   1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1; 1;
@@ -95,38 +84,39 @@ let utf_8_len = [| (* uchar byte length according to first UTF-8 byte. *)
 
 let r_utf_8 s j l =
   (* assert (0 <= j && 0 <= l && j + l <= String.length s); *)
+  let uchar c = `Uchar (Uchar.unsafe_of_int c) in
   match l with
-  | 1 -> `Uchar (unsafe_byte s j)
+  | 1 -> uchar (unsafe_byte s j)
   | 2 ->
       let b0 = unsafe_byte s j in let b1 = unsafe_byte s (j + 1) in
       if b1 lsr 6 != 0b10 then malformed s j l else
-      `Uchar (((b0 land 0x1F) lsl 6) lor (b1 land 0x3F))
+      uchar (((b0 land 0x1F) lsl 6) lor (b1 land 0x3F))
   | 3 ->
       let b0 = unsafe_byte s j in let b1 = unsafe_byte s (j + 1) in
       let b2 = unsafe_byte s (j + 2) in
-      let c = `Uchar (((b0 land 0x0F) lsl 12) lor
-                      ((b1 land 0x3F) lsl 6) lor
-          (b2 land 0x3F))
+      let c = ((b0 land 0x0F) lsl 12) lor
+              ((b1 land 0x3F) lsl 6) lor
+              (b2 land 0x3F)
       in
       if b2 lsr 6 != 0b10 then malformed s j l else
       begin match b0 with
-      | 0xE0 -> if b1 < 0xA0 || 0xBF < b1 then malformed s j l else c
-      | 0xED -> if b1 < 0x80 || 0x9F < b1 then malformed s j l else c
-      | _ -> if b1 lsr 6 != 0b10 then malformed s j l else c
+      | 0xE0 -> if b1 < 0xA0 || 0xBF < b1 then malformed s j l else uchar c
+      | 0xED -> if b1 < 0x80 || 0x9F < b1 then malformed s j l else uchar c
+      | _ -> if b1 lsr 6 != 0b10 then malformed s j l else uchar c
       end
   | 4 ->
       let b0 = unsafe_byte s j in let b1 = unsafe_byte s (j + 1) in
       let b2 = unsafe_byte s (j + 2) in let b3 = unsafe_byte s (j + 3) in
-      let c = `Uchar (((b0 land 0x07) lsl 18) lor
-                      ((b1 land 0x3F) lsl 12) lor
-          ((b2 land 0x3F) lsl 6) lor
-                      (b3 land 0x3F))
+      let c = (((b0 land 0x07) lsl 18) lor
+               ((b1 land 0x3F) lsl 12) lor
+               ((b2 land 0x3F) lsl 6) lor
+               (b3 land 0x3F))
       in
       if b3 lsr 6 != 0b10 || b2 lsr 6 != 0b10 then malformed s j l else
       begin match b0 with
-      | 0xF0 -> if b1 < 0x90 || 0xBF < b1 then malformed s j l else c
-      | 0xF4 -> if b1 < 0x80 || 0x8F < b1 then malformed s j l else c
-      | _ -> if b1 lsr 6 != 0b10 then malformed s j l else c
+      | 0xF0 -> if b1 < 0x90 || 0xBF < b1 then malformed s j l else uchar c
+      | 0xF4 -> if b1 < 0x80 || 0x8F < b1 then malformed s j l else uchar c
+      | _ -> if b1 lsr 6 != 0b10 then malformed s j l else uchar c
       end
   | _ -> assert false
 
@@ -134,7 +124,7 @@ let r_utf_16 s j0 j1 =                       (* May return a high surrogate. *)
   (* assert (0 <= j0 && 0 <= j1 && max j0 j1 < String.length s); *)
   let b0 = unsafe_byte s j0 in let b1 = unsafe_byte s j1 in
   let u = (b0 lsl 8) lor b1 in
-  if u < 0xD800 || u > 0xDFFF then `Uchar u else
+  if u < 0xD800 || u > 0xDFFF then `Uchar (Uchar.unsafe_of_int u) else
   if u > 0xDBFF then malformed s (min j0 j1) 2 else `Hi u
 
 let r_utf_16_lo hi s j0 j1 =          (* Combines [hi] with a low surrogate. *)
@@ -144,7 +134,8 @@ let r_utf_16_lo hi s j0 j1 =          (* Combines [hi] with a low surrogate. *)
   let lo = (b0 lsl 8) lor b1 in
   if lo < 0xDC00 || lo > 0xDFFF
   then malformed_pair (j0 < j1 (* true => be *)) hi s (min j0 j1) 2
-  else `Uchar ((((hi land 0x3FF) lsl 10) lor (lo land 0x3FF)) + 0x10000)
+  else `Uchar (Uchar.unsafe_of_int ((((hi land 0x3FF) lsl 10) lor
+                                     (lo land 0x3FF)) + 0x10000))
 
 let r_encoding s j l =                  (* guess encoding with max. 3 bytes. *)
   (* assert (0 <= j && 0 <= l && j + l <= String.length s) *)
@@ -166,11 +157,11 @@ let r_encoding s j l =                  (* guess encoding with max. 3 bytes. *)
 (* Decode *)
 
 type src = [ `Channel of in_channel | `String of string | `Manual ]
-type nln = [ `ASCII of uchar | `NLF of uchar | `Readline of uchar ]
-type decode = [ `Await | `End | `Malformed of string | `Uchar of uchar]
+type nln = [ `ASCII of Uchar.t | `NLF of Uchar.t | `Readline of Uchar.t ]
+type decode = [ `Await | `End | `Malformed of string | `Uchar of Uchar.t]
 
 let pp_decode ppf = function
-| `Uchar u -> pp ppf "@[`Uchar %a@]" pp_cp u
+| `Uchar u -> pp ppf "@[`Uchar U+%04X@]" (Uchar.to_int u)
 | `End -> pp ppf "`End"
 | `Await -> pp ppf "`Await"
 | `Malformed bs ->
@@ -184,11 +175,11 @@ type decoder =
   { src : src;                                              (* input source. *)
     mutable encoding : decoder_encoding;                (* decoded encoding. *)
     nln : nln option;                     (* newline normalization (if any). *)
-    nl : int;                            (* newline normalization character. *)
-    mutable i : string;                              (* current input chunk. *)
+    nl : Uchar.t;                        (* newline normalization character. *)
+    mutable i : Bytes.t;                             (* current input chunk. *)
     mutable i_pos : int;                          (* input current position. *)
     mutable i_max : int;                          (* input maximal position. *)
-    t : string;        (* four bytes temporary buffer for overlapping reads. *)
+    t : Bytes.t;       (* four bytes temporary buffer for overlapping reads. *)
     mutable t_len : int;                      (* current byte length of [t]. *)
     mutable t_need : int;                  (* number of bytes needed in [t]. *)
     mutable removed_bom : bool;     (* [true] if an initial BOM was removed. *)
@@ -198,7 +189,7 @@ type decoder =
     mutable byte_count : int;                                 (* byte count. *)
     mutable count : int;                                      (* char count. *)
     mutable pp :        (* decoder post-processor for BOM, position and nln. *)
-      decoder -> [ `Malformed of string | `Uchar of uchar ] -> decode;
+      decoder -> [ `Malformed of string | `Uchar of Uchar.t ] -> decode;
     mutable k : decoder -> decode }                 (* decoder continuation. *)
 
 (* On decodes that overlap two (or more) [d.i] buffers, we use [t_fill] to copy
@@ -209,9 +200,11 @@ type decoder =
    which implies that [i_rem d < 0] is [true]. *)
 
 let i_rem d = d.i_max - d.i_pos + 1     (* remaining bytes to read in [d.i]. *)
-let eoi d = d.i <- ""; d.i_pos <- 0; d.i_max <- min_int   (* set eoi in [d]. *)
+let eoi d =
+  d.i <- Bytes.empty; d.i_pos <- 0; d.i_max <- min_int   (* set eoi in [d]. *)
+
 let src d s j l =                                     (* set [d.i] with [s]. *)
-  if (j < 0 || l < 0 || j + l > String.length s) then invalid_bounds j l else
+  if (j < 0 || l < 0 || j + l > Bytes.length s) then invalid_bounds j l else
   if (l = 0) then eoi d else
   (d.i <- s; d.i_pos <- j; d.i_max <- j + l - 1)
 
@@ -219,7 +212,7 @@ let refill k d = match d.src with  (* get new input in [d.i] and [k]ontinue. *)
 | `Manual -> d.k <- k; `Await
 | `String _ -> eoi d; k d
 | `Channel ic ->
-    let rc = input ic d.i 0 (String.length d.i) in
+    let rc = input ic d.i 0 (Bytes.length d.i) in
     (src d d.i 0 rc; k d)
 
 let t_need d need = d.t_len <- 0; d.t_need <- need
@@ -385,13 +378,13 @@ let guessed_utf_16 d be v =     (* start decoder after `UTF_16{BE,LE} guess. *)
   in
   match v with
   | `BOM -> ret (b3 t_decode_utf_16) (`Uchar u_bom) 2 d
-  | `ASCII u -> ret (b3 t_decode_utf_16) (`Uchar u) 2 d
+  | `ASCII u -> ret (b3 t_decode_utf_16) (`Uchar (Uchar.unsafe_of_int u)) 2 d
   | `Decode ->
       match r_utf_16 d.t j0 j1 with
       | `Malformed _ | `Uchar _ as v -> ret (b3 t_decode_utf_16) v 2 d
       | `Hi hi ->
         if d.t_len < 3
-        then ret decode_utf_16 (malformed_pair be hi "" 0 0) d.t_len d
+        then ret decode_utf_16 (malformed_pair be hi Bytes.empty 0 0) d.t_len d
         else (b3 (t_decode_utf_16_lo hi)) d
 
 let guess_encoding d =                  (* guess encoding and start decoder. *)
@@ -422,51 +415,70 @@ let ncount d = d.count <- d.count + 1                            (* inlined. *)
 let cr d b = d.last_cr <- b                                      (* inlined. *)
 
 let pp_remove_bom utf16 pp d = function(* removes init. BOM, handles UTF-16. *)
-| `Uchar 0xFEFF (* BOM *) ->
-    if utf16 then (d.encoding <- `UTF_16BE; d.k <- decode_utf_16be);
-    d.removed_bom <- true; d.pp <- pp; d.k d
-| `Uchar 0xFFFE (* BOM reversed from decode_utf_16be *) when utf16 ->
-    d.encoding <- `UTF_16LE; d.k <- decode_utf_16le;
-    d.removed_bom <- true; d.pp <- pp; d.k d
-| `Malformed _ | `Uchar _ as v ->
-    d.removed_bom <- false; d.pp <- pp; d.pp d v
+| `Malformed _  as v -> d.removed_bom <- false; d.pp <- pp; d.pp d v
+| `Uchar u as v ->
+    match Uchar.to_int u with
+    | 0xFEFF (* BOM *) ->
+        if utf16 then (d.encoding <- `UTF_16BE; d.k <- decode_utf_16be);
+        d.removed_bom <- true; d.pp <- pp; d.k d
+    | 0xFFFE (* BOM reversed from decode_utf_16be *) when utf16 ->
+        d.encoding <- `UTF_16LE; d.k <- decode_utf_16le;
+        d.removed_bom <- true; d.pp <- pp; d.k d
+    | _ ->
+        d.removed_bom <- false; d.pp <- pp; d.pp d v
 
 let pp_nln_none d = function
-| `Uchar 0x000A (* LF *) as v ->
-    let last_cr = d.last_cr in
-    cr d false; ncount d; if last_cr then v else (nline d; v)
-| `Uchar 0x000D (* CR *) as v -> cr d true; ncount d; nline d; v
-| `Uchar (0x0085 | 0x000C | 0x2028 | 0x2029) (* NEL | FF | LS | PS *) as v ->
-    cr d false; ncount d; nline d; v
-| `Uchar _ | `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Uchar u as v ->
+    match Uchar.to_int u with
+    | 0x000A (* LF *) ->
+        let last_cr = d.last_cr in
+        cr d false; ncount d; if last_cr then v else (nline d; v)
+    | 0x000D (* CR *) -> cr d true; ncount d; nline d; v
+    | (0x0085 | 0x000C | 0x2028 | 0x2029) (* NEL | FF | LS | PS *) ->
+        cr d false; ncount d; nline d; v
+    | _ ->
+        cr d false; ncount d; ncol d; v
 
 let pp_nln_readline d = function
-| `Uchar 0x000A (* LF *) ->
-    let last_cr = d.last_cr in
-    cr d false; if last_cr then d.k d else (ncount d; nline d; `Uchar d.nl)
-| `Uchar 0x000D (* CR *) -> cr d true; ncount d; nline d; `Uchar d.nl
-| `Uchar (0x0085 | 0x000C | 0x2028 | 0x2029) (* NEL | FF | LS | PS *) ->
-    cr d false; ncount d; nline d; `Uchar d.nl
-| `Uchar _ | `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Uchar u as v ->
+    match Uchar.to_int u with
+    | 0x000A (* LF *) ->
+        let last_cr = d.last_cr in
+        cr d false; if last_cr then d.k d else (ncount d; nline d; `Uchar d.nl)
+    | 0x000D (* CR *) -> cr d true; ncount d; nline d; `Uchar d.nl
+    | (0x0085 | 0x000C | 0x2028 | 0x2029) (* NEL | FF | LS | PS *) ->
+        cr d false; ncount d; nline d; `Uchar d.nl
+    | _ ->
+        cr d false; ncount d; ncol d; v
 
 let pp_nln_nlf d = function
-| `Uchar 0x000A (* LF *) ->
-    let last_cr = d.last_cr in
-    cr d false; if last_cr then d.k d else (ncount d; nline d; `Uchar d.nl)
-| `Uchar 0x000D (* CR *) -> cr d true; ncount d; nline d; `Uchar d.nl
-| `Uchar 0x0085 (* NEL *) -> cr d false; ncount d; nline d; `Uchar d.nl
-| `Uchar (0x000C | 0x2028 | 0x2029) as v (* FF | LS | PS *) ->
-    cr d false; ncount d; nline d; v
-| `Uchar _ | `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Uchar u as v ->
+    match Uchar.to_int u with
+    | 0x000A (* LF *) ->
+        let last_cr = d.last_cr in
+        cr d false; if last_cr then d.k d else (ncount d; nline d; `Uchar d.nl)
+    | 0x000D (* CR *) -> cr d true; ncount d; nline d; `Uchar d.nl
+    | 0x0085 (* NEL *) -> cr d false; ncount d; nline d; `Uchar d.nl
+    | (0x000C | 0x2028 | 0x2029) (* FF | LS | PS *) ->
+        cr d false; ncount d; nline d; v
+    | _ ->
+        cr d false; ncount d; ncol d; v
 
 let pp_nln_ascii d = function
-| `Uchar 0x000A (* LF *) ->
-    let last_cr = d.last_cr in
-    cr d false; if last_cr then d.k d else (ncount d; nline d; `Uchar d.nl)
-| `Uchar 0x000D (* CR *) -> cr d true; ncount d; nline d; `Uchar d.nl
-| `Uchar (0x0085 | 0x000C | 0x2028 | 0x2029) as v (* NEL | FF | LS | PS *) ->
-    cr d false; ncount d; nline d; v
-| `Uchar _ | `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Malformed _ as v -> cr d false; ncount d; ncol d; v
+| `Uchar u as v ->
+    match Uchar.to_int u with
+    | 0x000A (* LF *) ->
+        let last_cr = d.last_cr in
+        cr d false; if last_cr then d.k d else (ncount d; nline d; `Uchar d.nl)
+    | 0x000D (* CR *) -> cr d true; ncount d; nline d; `Uchar d.nl
+    | (0x0085 | 0x000C | 0x2028 | 0x2029) (* NEL | FF | LS | PS *) ->
+        cr d false; ncount d; nline d; v
+    | _ ->
+        cr d false; ncount d; ncol d; v
 
 let decode_fun = function
 | `UTF_8 -> decode_utf_8
@@ -478,7 +490,7 @@ let decode_fun = function
 
 let decoder ?nln ?encoding src =
   let pp, nl = match nln with
-  | None -> pp_nln_none, 0x000A (* not used. *)
+  | None -> pp_nln_none, Uchar.unsafe_of_int 0x000A (* not used. *)
   | Some (`ASCII nl) -> pp_nln_ascii, nl
   | Some (`NLF nl) -> pp_nln_nlf, nl
   | Some (`Readline nl) -> pp_nln_readline, nl
@@ -488,12 +500,12 @@ let decoder ?nln ?encoding src =
   | Some e -> (e :> decoder_encoding), decode_fun e
   in
   let i, i_pos, i_max = match src with
-  | `Manual -> "", 1, 0                            (* implies src_rem d = 0. *)
-  | `Channel _ -> String.create io_buffer_size, 1, 0                (* idem. *)
-  | `String s -> s, 0, String.length s - 1
+  | `Manual -> Bytes.empty, 1, 0                  (* implies src_rem d = 0. *)
+  | `Channel _ -> Bytes.create io_buffer_size, 1, 0                (* idem. *)
+  | `String s -> Bytes.unsafe_of_string s, 0, String.length s - 1
   in
   { src = (src :> src); encoding; nln = (nln :> nln option); nl;
-    i; i_pos; i_max; t = String.create 4; t_len = 0; t_need = 0;
+    i; i_pos; i_max; t = Bytes.create 4; t_len = 0; t_need = 0;
     removed_bom = false; last_cr = false; line = 1; col = 0;
     byte_count = 0; count = 0;
     pp = pp_remove_bom (encoding = `UTF_16) pp; k }
@@ -513,14 +525,14 @@ let set_decoder_encoding d e =
 (* Encode *)
 
 type dst = [ `Channel of out_channel | `Buffer of Buffer.t | `Manual ]
-type encode = [ `Await | `End | `Uchar of uchar ]
+type encode = [ `Await | `End | `Uchar of Uchar.t ]
 type encoder =
   { dst : dst;                                        (* output destination. *)
     encoding : encoding;                                (* encoded encoding. *)
-    mutable o : string;                             (* current output chunk. *)
+    mutable o : Bytes.t;                            (* current output chunk. *)
     mutable o_pos : int;                   (* next output position to write. *)
     mutable o_max : int;                (* maximal output position to write. *)
-    t : string;                 (* four bytes buffer for overlapping writes. *)
+    t : Bytes.t;                (* four bytes buffer for overlapping writes. *)
     mutable t_pos : int;                    (* next position to read in [t]. *)
     mutable t_max : int;                 (* maximal position to read in [t]. *)
     mutable k :                                     (* encoder continuation. *)
@@ -534,14 +546,17 @@ type encoder =
 
 let o_rem e = e.o_max - e.o_pos + 1    (* remaining bytes to write in [e.o]. *)
 let dst e s j l =                                     (* set [e.o] with [s]. *)
-  if (j < 0 || l < 0 || j + l > String.length s) then invalid_bounds j l;
+  if (j < 0 || l < 0 || j + l > Bytes.length s) then invalid_bounds j l;
   e.o <- s; e.o_pos <- j; e.o_max <- j + l - 1
 
 let partial k e = function `Await -> k e | `Uchar _ | `End -> invalid_encode ()
 let flush k e = match e.dst with(* get free storage in [d.o] and [k]ontinue. *)
 | `Manual -> e.k <- partial k; `Partial
-| `Buffer b -> Buffer.add_substring b e.o 0 e.o_pos; e.o_pos <- 0; k e
 | `Channel oc -> output oc e.o 0 e.o_pos; e.o_pos <- 0; k e
+| `Buffer b ->
+    let o = Bytes.unsafe_to_string e.o in
+    Buffer.add_substring b o 0 e.o_pos; e.o_pos <- 0; k e
+
 
 let t_range e max = e.t_pos <- 0; e.t_max <- max
 let rec t_flush k e =               (* flush [d.t] up to [d.t_max] in [d.i]. *)
@@ -561,6 +576,7 @@ let rec encode_utf_8 e v =
   | `Await -> k e
   | `End -> flush k e
   | `Uchar u as v ->
+      let u = Uchar.to_int u in
       let rem = o_rem e in
       if u <= 0x007F then
       if rem < 1 then flush (fun e -> encode_utf_8 e v) e else
@@ -605,6 +621,7 @@ let rec encode_utf_16be e v =
   | `Await -> k e
   | `End -> flush k e
   | `Uchar u ->
+      let u = Uchar.to_int u in
       let rem = o_rem e in
       if u < 0x10000 then
       begin
@@ -636,32 +653,33 @@ let rec encode_utf_16le e v =         (* encode_uft_16be with bytes swapped. *)
   | `Await -> k e
   | `End -> flush k e
   | `Uchar u ->
-    let rem = o_rem e in
-    if u < 0x10000 then
-    begin
-      let s, j, k =
-        if rem < 2 then (t_range e 1; e.t, 0, t_flush k) else
-        let j = e.o_pos in (e.o_pos <- e.o_pos + 2; e.o, j, k)
-      in
-      unsafe_set_byte s j (u land 0xFF);
-      unsafe_set_byte s (j + 1) (u lsr 8);
-      k e
-    end
-    else
-    begin
-      let s, j, k =
-        if rem < 4 then (t_range e 3; e.t, 0, t_flush k) else
-        let j = e.o_pos in (e.o_pos <- e.o_pos + 4; e.o, j, k)
-      in
-      let u' = u - 0x10000 in
-      let hi = (0xD800 lor (u' lsr 10)) in
-      let lo = (0xDC00 lor (u' land 0x3FF)) in
-      unsafe_set_byte s j (hi land 0xFF);
-      unsafe_set_byte s (j + 1) (hi lsr 8);
-      unsafe_set_byte s (j + 2) (lo land 0xFF);
-      unsafe_set_byte s (j + 3) (lo lsr 8);
-      k e
-    end
+      let u = Uchar.to_int u in
+      let rem = o_rem e in
+      if u < 0x10000 then
+        begin
+          let s, j, k =
+            if rem < 2 then (t_range e 1; e.t, 0, t_flush k) else
+            let j = e.o_pos in (e.o_pos <- e.o_pos + 2; e.o, j, k)
+          in
+          unsafe_set_byte s j (u land 0xFF);
+          unsafe_set_byte s (j + 1) (u lsr 8);
+          k e
+        end
+      else
+      begin
+        let s, j, k =
+          if rem < 4 then (t_range e 3; e.t, 0, t_flush k) else
+          let j = e.o_pos in (e.o_pos <- e.o_pos + 4; e.o, j, k)
+        in
+        let u' = u - 0x10000 in
+        let hi = (0xD800 lor (u' lsr 10)) in
+        let lo = (0xDC00 lor (u' land 0x3FF)) in
+        unsafe_set_byte s j (hi land 0xFF);
+        unsafe_set_byte s (j + 1) (hi lsr 8);
+        unsafe_set_byte s (j + 2) (lo land 0xFF);
+        unsafe_set_byte s (j + 3) (lo lsr 8);
+        k e
+      end
 
 let encode_fun = function
 | `UTF_8 -> encode_utf_8
@@ -671,12 +689,12 @@ let encode_fun = function
 
 let encoder encoding dst =
   let o, o_pos, o_max = match dst with
-  | `Manual -> "", 1, 0                              (* implies o_rem e = 0. *)
+  | `Manual -> Bytes.empty, 1, 0                     (* implies o_rem e = 0. *)
   | `Buffer _
-  | `Channel _ -> String.create io_buffer_size, 0, io_buffer_size - 1
+  | `Channel _ -> Bytes.create io_buffer_size, 0, io_buffer_size - 1
   in
   { dst = (dst :> dst); encoding = (encoding :> encoding); o; o_pos; o_max;
-    t = String.create 4; t_pos = 1; t_max = 0; k = encode_fun encoding}
+    t = Bytes.create 4; t_pos = 1; t_max = 0; k = encode_fun encoding}
 
 let encode e v = e.k e (v :> encode)
 let encoder_encoding e = e.encoding
@@ -693,54 +711,64 @@ end
 (* Strings folders and Buffer encoders *)
 
 module String = struct
-  let encoding_guess s = match r_encoding s 0 (max (String.length s) 3) with
-  | `UTF_8 d -> `UTF_8, (d = `BOM)
-  | `UTF_16BE d -> `UTF_16BE, (d = `BOM)
-  | `UTF_16LE d -> `UTF_16LE, (d = `BOM)
+  let encoding_guess s =
+    let s = Bytes.unsafe_of_string s in
+    match r_encoding s 0 (max (Bytes.length s) 3) with
+    | `UTF_8 d -> `UTF_8, (d = `BOM)
+    | `UTF_16BE d -> `UTF_16BE, (d = `BOM)
+    | `UTF_16LE d -> `UTF_16LE, (d = `BOM)
 
   type 'a folder =
-    'a -> int -> [ `Uchar of uchar | `Malformed of string ] -> 'a
+    'a -> int -> [ `Uchar of Uchar.t | `Malformed of string ] -> 'a
 
-  let fold_utf_8 f acc s =
-    let rec loop acc f s i l =
-      if i = l then acc else
+  let fold_utf_8 ?(pos = 0) ?len f acc s =
+    let rec loop acc f s i last =
+      if i > last then acc else
       let need = unsafe_array_get utf_8_len (unsafe_byte s i) in
-      if need = 0 then loop (f acc i (malformed s i 1)) f s (i + 1) l else
-      let rem = l - i in
+      if need = 0 then loop (f acc i (malformed s i 1)) f s (i + 1) last else
+      let rem = last - i + 1 in
       if rem < need then f acc i (malformed s i rem) else
-      loop (f acc i (r_utf_8 s i need)) f s (i + need) l
+      loop (f acc i (r_utf_8 s i need)) f s (i + need) last
     in
-    loop acc f s 0 (String.length s)
+    let len = match len with None -> String.length s - pos | Some l -> l in
+    let last = pos + len - 1 in
+    loop acc f (Bytes.unsafe_of_string s) pos last
 
-  let fold_utf_16be f acc s =
-    let rec loop acc f s i l =
-      if i = l then acc else
-      let rem = l - i in
+  let fold_utf_16be ?(pos = 0) ?len f acc s =
+    let rec loop acc f s i last =
+      if i > last then acc else
+      let rem = last - i + 1 in
       if rem < 2 then f acc i (malformed s i 1) else
       match r_utf_16 s i (i + 1) with
-      | `Uchar _ | `Malformed _ as v -> loop (f acc i v) f s (i + 2) l
+      | `Uchar _ | `Malformed _ as v -> loop (f acc i v) f s (i + 2) last
       | `Hi hi ->
           if rem < 4 then f acc i (malformed s i rem)  else
-          loop (f acc i (r_utf_16_lo hi s (i + 2) (i + 3))) f s (i + 4) l
+          loop (f acc i (r_utf_16_lo hi s (i + 2) (i + 3))) f s (i + 4) last
     in
-    loop acc f s 0 (String.length s)
+    let len = match len with None -> String.length s - pos | Some l -> l in
+    let last = pos + len - 1 in
+    loop acc f (Bytes.unsafe_of_string s) pos last
 
-  let fold_utf_16le f acc s =             (* [fold_utf_16be], bytes swapped. *)
-    let rec loop acc f s i l =
-      if i = l then acc else
-      let rem = l - i in
+  let fold_utf_16le ?(pos = 0) ?len f acc s =
+    (* [fold_utf_16be], bytes swapped. *)
+    let rec loop acc f s i last =
+      if i > last then acc else
+      let rem = last - i + 1 in
       if rem < 2 then f acc i (malformed s i 1) else
       match r_utf_16 s (i + 1) i with
-      | `Uchar _ | `Malformed _ as v -> loop (f acc i v) f s (i + 2) l
+      | `Uchar _ | `Malformed _ as v -> loop (f acc i v) f s (i + 2) last
       | `Hi hi ->
           if rem < 4 then f acc i (malformed s i rem)  else
-          loop (f acc i (r_utf_16_lo hi s (i + 3) (i + 2))) f s (i + 4) l
+          loop (f acc i (r_utf_16_lo hi s (i + 3) (i + 2))) f s (i + 4) last
     in
-    loop acc f s 0 (String.length s)
+    let len = match len with None -> String.length s - pos | Some l -> l in
+    let last = pos + len - 1 in
+    loop acc f (Bytes.unsafe_of_string s) pos last
 end
 
 module Buffer = struct
   let add_utf_8 b u =
+    let u = Uchar.to_int u in
     let w byte = Buffer.add_char b (unsafe_chr byte) in          (* inlined. *)
     if u <= 0x007F then
     (w u)
@@ -758,6 +786,7 @@ module Buffer = struct
      w (0x80 lor (u land 0x3F)))
 
   let add_utf_16be b u =
+    let u = Uchar.to_int u in
     let w byte = Buffer.add_char b (unsafe_chr byte) in          (* inlined. *)
     if u < 0x10000 then (w (u lsr 8); w (u land 0xFF)) else
     let u' = u - 0x10000 in
@@ -767,6 +796,7 @@ module Buffer = struct
     w (lo lsr 8); w (lo land 0xFF)
 
   let add_utf_16le b u =                            (* swapped add_utf_16be. *)
+    let u = Uchar.to_int u in
     let w byte = Buffer.add_char b (unsafe_chr byte) in          (* inlined. *)
     if u < 0x10000 then (w (u land 0xFF); w (u lsr 8)) else
     let u' = u - 0x10000 in
@@ -777,34 +807,17 @@ module Buffer = struct
 end
 
 (*---------------------------------------------------------------------------
-   Copyright 2012 Daniel C. Bünzli
-   All rights reserved.
+   Copyright (c) 2012 Daniel C. Bünzli
 
-   Redistribution and use in source and binary forms, with or without
-   modification, are permitted provided that the following conditions
-   are met:
+   Permission to use, copy, modify, and/or distribute this software for any
+   purpose with or without fee is hereby granted, provided that the above
+   copyright notice and this permission notice appear in all copies.
 
-   1. Redistributions of source code must retain the above copyright
-      notice, this list of conditions and the following disclaimer.
-
-   2. Redistributions in binary form must reproduce the above
-      copyright notice, this list of conditions and the following
-      disclaimer in the documentation and/or other materials provided
-      with the distribution.
-
-   3. Neither the name of Daniel C. Bünzli nor the names of
-      contributors may be used to endorse or promote products derived
-      from this software without specific prior written permission.
-
-   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+   THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+   WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+   MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+   ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+   ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
   ---------------------------------------------------------------------------*)
